use std::time::Instant;
use log::info;
use noise::{Fbm, MultiFractal, NoiseFn, Perlin};
use rand::prelude::*;
use voronoice::{BoundingBox, Point, Voronoi, VoronoiBuilder};
use winit::{event::{DeviceEvent, ElementState, Event, KeyEvent, MouseButton, MouseScrollDelta, RawKeyEvent, WindowEvent}, keyboard::{KeyCode, PhysicalKey}, window::Window};

use crate::graphics::{Uniforms, Vertex};

const FRAMERATE: usize = 10; // Update per second

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum CellKind {
    Void,
    Sea,
    Beach,
    Forest,
    Dirt,
    Stone,
    Grass
}

struct CellData {
    kind: CellKind,
    cell: usize,
    z: f32,
    moisture: f32
}
impl CellData {
    fn new(kind: CellKind, cell: usize, z: f32, moisture: f32) -> Self {
        Self {
            kind,
            cell,
            z,
            moisture
        }
    }
    fn color(&self) -> [f32; 4] {
        // let mut rng = thread_rng();
        // [rng.gen(), rng.gen(), rng.gen(), 1.]
        match self.kind {
            CellKind::Void => [0.; 4],
            CellKind::Sea => [0., 0., 1., 1.],
            CellKind::Beach => [0.82, 0.84, 0.51, 1.],
            CellKind::Forest => [0., 0.5 - (self.moisture*0.4), 0., 1.],
            CellKind::Dirt => [0.53 - (self.moisture*0.4), 0.38-(self.moisture*0.4), 0.29-(self.moisture*0.4), 1.],
            CellKind::Stone => [0.5, 0.5, 0.5, 1.],
            CellKind::Grass => [(136./255.) - (self.moisture*0.4), (204./255.) - (self.moisture*0.4), (59./255.) - (self.moisture*0.4), 1.]
        }
    }
}

struct Map {
    voronoi: Voronoi,
    cells_data: Vec<CellData>,
    seed: u32
}
impl Map {
    const HEIGHT: f32 = 2.;
    const WIDTH: f32 = 2.;
    const SIZE: usize = 10_000;
    fn new(seed: u32) -> Self {
        let mut rng = rand::rngs::SmallRng::seed_from_u64(seed as u64);
        let mut sites = Vec::with_capacity(Self::SIZE);
        for _ in 0..Self::SIZE {
            sites.push(Point { x:rng.gen_range(-Self::WIDTH/2.0..Self::WIDTH/2.0) as f64, y:rng.gen_range(-Self::HEIGHT/2.0..Self::HEIGHT/2.0) as f64 })
        }
        let voronoi = VoronoiBuilder::default()
            .set_sites(sites)
            .set_bounding_box(BoundingBox::new_centered(Self::WIDTH as f64, Self::HEIGHT as f64))
            .set_lloyd_relaxation_iterations(3)
            .build()
            .unwrap();
        let mut cells_data = Vec::with_capacity(Self::SIZE);
        let z_noise = Fbm::<Perlin>::new(seed);
        let moisture_noise = Fbm::<Perlin>::new(seed+1)
            .set_frequency(2.);
        for i in 0..Self::SIZE {
            let c = voronoi.cell(i);
            let site = c.site_position();
            let z = (
                0.3 // Arbitrary value
                + ((z_noise.get([site.x, site.y])+1.)/2.) // Noise + [0; 1]
                - ((site.x.powi(2)+site.y.powi(2)).sqrt()*0.5) // Distance - [0; sqrt(2)] * 0.5
            ).clamp(0., 1.);
            let m = (
                (moisture_noise.get([site.x, site.y])+1.)/2. // Noise + [0; 1]
            ).clamp(0., 1.) as f32;
            let k = if z <= 0.5 {
                CellKind::Sea
            } else if z <= 0.52 {
                CellKind::Beach
            } else if z < 0.8 {
                CellKind::Dirt
            } else {
                CellKind::Stone
            };
            cells_data.push(CellData::new(k, i, z as f32, m));
        }
        Self {
            voronoi,
            cells_data,
            seed
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum EntityKind {
    Horse
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum EntityState {
    Walking(Instant), // Start of walk
    Resting
}

#[derive(Debug)]
struct Entity {
    pos: [f32; 3],
    kind: EntityKind,
    start: Instant,
    state: EntityState
}
impl Entity {
    fn new(pos: [f32; 3], kind: EntityKind) -> Self {
        Self {
            pos,
            kind,
            start: Instant::now(),
            state: EntityState::Resting
        }
    }
    fn render(&self, vertices: &mut Vec<Vertex>, indices: &mut Vec<u32>) {
        match self.kind {
            EntityKind::Horse => {
                let color = [171./255., 122./255., 50./255., 1.];
                let dark = [color[0]-0.3, color[1]-0.3, color[2]-0.3, 1.];
                let (vs, is) = match self.state {
                    EntityState::Walking(now) => {
                        let now = now.elapsed().as_secs_f32()*5.;
                        (
                            [
                                // back left leg
                                Vertex { pos: [-0.5, 0.3], color: dark },
                                Vertex { pos: [-0.4 + (now.sin()*0.1), -0.7 + (now.cos().max(-0.5)*0.1)], color: dark },
                                Vertex { pos: [-0.25, 0.1], color: dark },

                                // back right leg
                                Vertex { pos: [-0.5, 0.3], color },
                                Vertex { pos: [-0.4 + ((now + 1.).sin()*0.1), -0.7 + ((now + 1.).cos().max(-0.5)*0.1)], color },
                                Vertex { pos: [-0.25, 0.1], color },

                                // front left leg
                                Vertex { pos: [0.3, 0.2], color: dark },
                                Vertex { pos: [0.4 + ((now-1.).sin()*0.1), -0.7 + ((now-1.).cos().max(-0.5)*0.1)], color: dark },
                                Vertex { pos: [0.5, 0.3], color: dark },

                                // front right leg
                                Vertex { pos: [0.3, 0.2], color },
                                Vertex { pos: [0.4 + ((now-2.).sin()*0.1), -0.7 + ((now-2.).cos().max(-0.5)*0.1)], color },
                                Vertex { pos: [0.5, 0.3], color },

                                // body
                                // 3
                                Vertex { pos: [-0.3, 0.], color },
                                Vertex { pos: [0.4, -0.1], color },
                                // 11
                                Vertex { pos: [0.3, 0.4], color },
                            ],
                            [
                                0,1,2,
                                3,4,5,
                                6,7,8,
                                9,10,11,
                                3,12,13,
                                3,13,11,
                                3,11,14
                            ]
                        )
                    },
                    EntityState::Resting => {
                        (
                            [
                                // back left leg
                                Vertex { pos: [-0.5, 0.3], color: dark },
                                Vertex { pos: [-0.4, -0.75], color: dark },
                                Vertex { pos: [-0.25, 0.1], color: dark },

                                // back right leg
                                Vertex { pos: [-0.5, 0.3], color },
                                Vertex { pos: [-0.4, -0.75], color },
                                Vertex { pos: [-0.25, 0.1], color },

                                // front left leg
                                Vertex { pos: [0.3, 0.2], color: dark },
                                Vertex { pos: [0.4, -0.75], color: dark },
                                Vertex { pos: [0.5, 0.3], color: dark },

                                // front right leg
                                Vertex { pos: [0.3, 0.2], color },
                                Vertex { pos: [0.4, -0.75], color },
                                Vertex { pos: [0.5, 0.3], color },

                                // body
                                // 3
                                Vertex { pos: [-0.3, 0.], color },
                                Vertex { pos: [0.4, -0.1], color },
                                // 11
                                Vertex { pos: [0.3, 0.4], color },
                            ],
                            [
                                0,1,2,
                                3,4,5,
                                6,7,8,
                                9,10,11,
                                3,12,13,
                                3,13,11,
                                3,11,14
                            ]
                        )
                    }
                };
                

                vertices.reserve(vs.len());
                let base = vertices.len() as u32;
                for mut v in vs {
                    v.pos[0] = v.pos[0]/50. + self.pos[0];
                    v.pos[1] = (v.pos[1] + 0.75)/50. + self.pos[1];
                    vertices.push(v)
                }

                indices.reserve(is.len());
                for i in is {
                    indices.push(base + i);
                }

                // let vs = vec![
                //     Vertex { pos: [-0.5, 0.3, 0.], color },
                //     Vertex { pos: [-0.3, 0., 0.], color },
                //     Vertex { pos: [0.4, -0.1, 0.], color },
                //     Vertex { pos: [0.5, 0.3, 0.], color },
                //     Vertex { pos: [0.3, 0.4, 0.], color },
                // ];
                // let i = self.vertices.len() as u32;
                // for v in vs.iter() {
                //     self.vertices.push(*v);
                // }
                // for v in 1..(vs.len()-1) as u32 {
                //     self.indices.push(i);
                //     self.indices.push(i+v);
                //     self.indices.push(i+v+1);
                // }
            }
        }
    }
}

pub struct State {
    pub vertices: Vec<Vertex>,
    pub indices: Vec<u32>,
    pub uniforms: Uniforms,
    map: Map,
    entities: Vec<Entity>,
    start: Instant,
    frame: usize,
    selected_tile: usize,
    mouse_pressed: bool
}
impl State {
    pub fn new() -> Self {
        let mut s = Self {
            vertices: vec![],
            indices: vec![],
            uniforms: Uniforms::default(),
            start: Instant::now(),
            frame: 0,
            map: Map::new(0),
            entities: Vec::new(),
            selected_tile: 0,
            mouse_pressed: false
        };
        s.render();
        s
    }
    pub fn event(&mut self, event: &Event<()>, window: &Window) {
        match event {
            Event::WindowEvent { event: WindowEvent::MouseInput { state, button, ..}, ..} => {
                if state.is_pressed() {
                    match button {
                        MouseButton::Left => {
                            self.mouse_pressed = true;
                        },
                        MouseButton::Right => self.map = Map::new(self.map.seed + 1),
                        _ => {}
                    };
                } else {
                    match button {
                        MouseButton::Left => {
                            self.mouse_pressed = false;
                        },
                        _ => {}
                    }
                }
            },
            Event::WindowEvent { event: WindowEvent::CursorMoved { position, .. }, ..} => {
                let w_size = window.inner_size();
                let pos = Point {
                    x: ((2.*position.x/(w_size.width as f64))-1.)*(self.uniforms.camera[2] as f64) + self.uniforms.camera[0] as f64,
                    y: (1.-(2.*position.y/(w_size.height as f64)))*(self.uniforms.camera[2] as f64) + self.uniforms.camera[1] as f64
                };
                let c = self.map.voronoi.cell(self.selected_tile);
                if self.mouse_pressed {
                    for i in c.iter_path(pos.clone()) {
                        self.selected_tile = i;
                        let cd = &mut self.map.cells_data[self.selected_tile];
                        if let CellKind::Dirt = cd.kind {
                            cd.kind = CellKind::Forest;
                        }

                        self.entities.push(Entity::new([pos.x as f32, pos.y as f32, 0.9], EntityKind::Horse));
                    }
                } else {
                    self.selected_tile = c.iter_path(pos).last().unwrap();
                }
            },
            Event::DeviceEvent { event: DeviceEvent::MouseWheel { delta }, ..} => {
                self.uniforms.camera[2] -= match delta {
                    MouseScrollDelta::PixelDelta(pos) => pos.y as f32,
                    MouseScrollDelta::LineDelta(_, y) => *y
                };
                self.uniforms.camera[2] = self.uniforms.camera[2].clamp(0.1, 1.);
            },
            Event::WindowEvent { event: WindowEvent::KeyboardInput { event: KeyEvent { physical_key: PhysicalKey::Code(kc), state, .. }, .. }, .. } => {
                if state.is_pressed() {
                    match kc {
                        KeyCode::KeyW => {
                            self.uniforms.camera[1] += 0.1;
                        },
                        KeyCode::KeyS => {
                            self.uniforms.camera[1] -= 0.1;
                        },
                        KeyCode::KeyA => {
                            self.uniforms.camera[0] -= 0.1;
                        },
                        KeyCode::KeyD => {
                            self.uniforms.camera[0] += 0.1;
                        },
                        _ => {}
                    }
                }
            },
            _ => {}
        }
    }
    pub fn render(&mut self) {
        info!("render");
        self.vertices = Vec::new();
        self.indices = Vec::new();

        for (c, cd) in self.map.voronoi.iter_cells().zip(self.map.cells_data.iter()).filter(|(_,cd)| cd.kind != CellKind::Forest) {
            let mut color = cd.color();
            if c.site() == self.selected_tile {
                color[0] = (color[0]+0.4).clamp(0., 1.);
                color[1] = (color[1]+0.4).clamp(0., 1.);
                color[2] = (color[2]+0.4).clamp(0., 1.);
            }
            let vs = c.iter_vertices().collect::<Vec<_>>();
            let i = self.vertices.len() as u32;
            for v in vs.iter() {
                self.vertices.push(Vertex { pos: [v.x as f32, v.y as f32], color });
            }
            for v in 1..(vs.len()-1) as u32 {
                self.indices.push(i);
                self.indices.push(i+v);
                self.indices.push(i+v+1);
            }
        }

        for e in self.entities.iter() {
            e.render(&mut self.vertices, &mut self.indices);
        }

        for (c, cd) in self.map.voronoi.iter_cells().zip(self.map.cells_data.iter()).filter(|(_,cd)| cd.kind == CellKind::Forest) {
            let mut color = cd.color();
            if c.site() == self.selected_tile {
                color[0] = (color[0]+0.4).clamp(0., 1.);
                color[1] = (color[1]+0.4).clamp(0., 1.);
                color[2] = (color[2]+0.4).clamp(0., 1.);
            }
            let vs = c.iter_vertices().collect::<Vec<_>>();
            let i = self.vertices.len() as u32;
            for v in vs.iter() {
                self.vertices.push(Vertex { pos: [v.x as f32, v.y as f32], color });
            }
            for v in 1..(vs.len()-1) as u32 {
                self.indices.push(i);
                self.indices.push(i+v);
                self.indices.push(i+v+1);
            }
        }
    }
    pub fn update_if_needed(&mut self) {
        while self.start.elapsed().as_secs_f32() > (self.frame as f32) / (FRAMERATE as f32) {
            self.update();
        }
    }
    pub fn update(&mut self) {
        info!("update");
        let mut rng = thread_rng();
        let mut new_kind = Vec::new();
        for cd in self.map.cells_data.iter() {
            if cd.kind == CellKind::Forest || cd.kind == CellKind::Grass {
                let r = rng.gen::<f32>();
                if r < (0.035*cd.moisture) {
                    let c = self.map.voronoi.cell(cd.cell);
                    let n = c.iter_neighbors().choose(&mut rng).unwrap();
                    let k = if r < (0.005*cd.moisture) && (self.map.cells_data[n].kind == CellKind::Dirt || self.map.cells_data[n].kind == CellKind::Grass) && cd.kind == CellKind::Forest {
                        Some(CellKind::Forest)
                    } else if self.map.cells_data[n].kind == CellKind::Dirt {
                        Some(CellKind::Grass)
                    } else {
                        None
                    };
                    if let Some(k) = k {
                        new_kind.push((n, k));
                    }
                }
            }
        }
        for (n, k) in new_kind {
            self.map.cells_data[n].kind = k;
        }

        self.frame += 1;
    }
}